"""
Package metadata
"""
__all__ = [
    '__project__',
    '__title__',
    '__author__',
    '__copyright__',
    '__license__',
    '__version__',
    '__mail__',
    '__maintainer__',
    '__status__',
    '__packagename__']

__project__ = 'ATHENA'
__title__ = "athena"
__author__ = "Marco Tezzele, Francesco Romor"
__copyright__ = "Copyright 2019-2021, Athena contributors"
__license__ = "MIT"
__version__ = "0.1.1"
__mail__ = 'marcotez@gmail.com, francesco.romor@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"
__packagename__ = "athena-mathlab"
